package com.lumberyard.graph.neo4j.plugins.communityclustering;

import java.util.Comparator;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.Set;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

public class DeltaModularityMatrix {
    private Map<Long, Map<Long, DeltaModularity>> deltaModularityTrees = Maps.newHashMap();
    private Map<Long, Queue<DeltaModularity>> deltaModularityMaxHeaps = Maps.newHashMap();
    private Map<Long, Double> fractionOfEdgeEndsAttachedToCommunity = Maps.newHashMap();

    public void addEdge(long node1Id, int node1Degree, long node2Id, int node2Degree, long totalEdgesInGraph) {
        if (deltaModularityTrees.get(node1Id) == null) {
            addNewRowForNode(node1Id);
        }

        if (deltaModularityTrees.get(node2Id) == null) {
            addNewRowForNode(node2Id);
        }

        double deltaModularityForJoiningNode1Node2 = Initializer.getInitialDeltaModularity(node1Degree, node2Degree,
                totalEdgesInGraph);
        DeltaModularity node1node2Delta = new DeltaModularity(node2Id, deltaModularityForJoiningNode1Node2);
        DeltaModularity node2node1Delta = new DeltaModularity(node1Id, deltaModularityForJoiningNode1Node2);

        deltaModularityMaxHeaps.get(node1Id).add(node1node2Delta);
        deltaModularityMaxHeaps.get(node2Id).add(node2node1Delta);
        deltaModularityTrees.get(node1Id).put(node2Id, node1node2Delta);
        deltaModularityTrees.get(node2Id).put(node1Id, node2node1Delta);
    }

    private void addNewRowForNode(long nodeId) {
        Map<Long, DeltaModularity> connectedNodes = Maps.newTreeMap();

        deltaModularityMaxHeaps.put(nodeId, new PriorityQueue<DeltaModularity>(1, new Comparator<DeltaModularity>() {
            @Override
            public int compare(DeltaModularity dmod1, DeltaModularity dmod2) {
                return -1 * Double.compare(dmod1.deltaModularity, dmod2.deltaModularity);
            }
        }));

        deltaModularityTrees.put(nodeId, connectedNodes);
    }

    public void joinCommunities(long community1Id, long community2Id) {
        updateDeltaModularityTreesForJoinedCommunities(community1Id, community2Id);

        /*
         * deltaModularityMaxHeaps.remove(community1Id);
         * 
         * for (long communityId : deltaModularityMaxHeaps.keySet()) { Map<Long,
         * DeltaModularity> deltaModularitiesForCommunity =
         * deltaModularityTrees.get(communityId); DeltaModularity
         * newDeltaModularityForJoiningCurrentCommunityWithCommunity2 = new
         * DeltaModularity(community2Id, 0.0); boolean update = false;
         * 
         * if (deltaModularitiesForCommunity.containsKey(community1Id) &&
         * deltaModularitiesForCommunity.containsKey(community2Id)) {
         * newDeltaModularityForJoiningCurrentCommunityWithCommunity2 = new
         * DeltaModularity( community2Id,
         * deltaModularitiesForCommunity.get(community1Id).deltaModularity +
         * deltaModularitiesForCommunity.get(community2Id).deltaModularity);
         * deltaModularityMaxHeaps
         * .get(communityId).remove(deltaModularitiesForCommunity
         * .get(community1Id)); deltaModularityMaxHeaps.get(communityId).remove(
         * deltaModularitiesForCommunity.get(community2Id)); update = true; }
         * else if (deltaModularitiesForCommunity.containsKey(community1Id)) {
         * newDeltaModularityForJoiningCurrentCommunityWithCommunity2 = new
         * DeltaModularity( community2Id,
         * deltaModularitiesForCommunity.get(community1Id).deltaModularity - 2 *
         * fractionOfEdgeEndsAttachedToCommunity.get(community2Id) *
         * fractionOfEdgeEndsAttachedToCommunity.get(communityId));
         * deltaModularityMaxHeaps
         * .get(communityId).remove(deltaModularitiesForCommunity
         * .get(community1Id)); update = true; } else if
         * (deltaModularitiesForCommunity.containsKey(community2Id)) {
         * newDeltaModularityForJoiningCurrentCommunityWithCommunity2 = new
         * DeltaModularity( community2Id,
         * deltaModularitiesForCommunity.get(community2Id).deltaModularity - 2 *
         * fractionOfEdgeEndsAttachedToCommunity.get(community1Id) *
         * fractionOfEdgeEndsAttachedToCommunity.get(communityId));
         * deltaModularityMaxHeaps
         * .get(communityId).remove(deltaModularitiesForCommunity
         * .get(community2Id)); update = true; }
         * 
         * if (update) { deltaModularityMaxHeaps.get(communityId).add(
         * newDeltaModularityForJoiningCurrentCommunityWithCommunity2); } }
         */

        fractionOfEdgeEndsAttachedToCommunity.put(community2Id, fractionOfEdgeEndsAttachedToCommunity.get(community1Id)
                + fractionOfEdgeEndsAttachedToCommunity.get(community2Id));

        fractionOfEdgeEndsAttachedToCommunity.put(community1Id, 0.0);
    }

    protected void updateDeltaModularityTreesForJoinedCommunities(long idOfCommunityToBeRemoved, long idOfCommunityToBeExpanded) {
        deltaModularityTrees.get(idOfCommunityToBeRemoved).remove(idOfCommunityToBeExpanded);
        deltaModularityTrees.get(idOfCommunityToBeExpanded).remove(idOfCommunityToBeRemoved);

        Set<DeltaModularity> otherCommunitiesConnectedToBothJoinedCommunities = Sets.intersection(
                Sets.newHashSet(deltaModularityTrees.get(idOfCommunityToBeRemoved).values()),
                Sets.newHashSet(deltaModularityTrees.get(idOfCommunityToBeExpanded).values()));
        Set<DeltaModularity> otherCommunitiesOnlyConnectedToCommunityToBeRemoved = Sets.difference(
                Sets.newHashSet(deltaModularityTrees.get(idOfCommunityToBeRemoved).values()),
                Sets.newHashSet(deltaModularityTrees.get(idOfCommunityToBeExpanded).values()));
        Set<DeltaModularity> otherCommunitiesOnlyConnectedToCommunityToBeExpanded = Sets.difference(
                Sets.newHashSet(deltaModularityTrees.get(idOfCommunityToBeExpanded).values()),
                Sets.newHashSet(deltaModularityTrees.get(idOfCommunityToBeRemoved).values()));

        for (DeltaModularity deltaModularityForOtherCommunity : otherCommunitiesConnectedToBothJoinedCommunities) {
            updateDeltaModularityTreesForCommunityAttachedToBothJoinedCommunities(deltaModularityForOtherCommunity.communityId,
                    idOfCommunityToBeRemoved, idOfCommunityToBeExpanded);
        }

        for (DeltaModularity deltaModularityForOtherCommunity : otherCommunitiesOnlyConnectedToCommunityToBeRemoved) {
            updateDeltaModularityTreesForCommunityAttachedToRemovedJoinedCommunity(
                    deltaModularityForOtherCommunity.communityId, idOfCommunityToBeRemoved, idOfCommunityToBeExpanded);
        }

        for (DeltaModularity deltaModularityForOtherCommunity : otherCommunitiesOnlyConnectedToCommunityToBeExpanded) {
            updateDeltaModularityTreesForCommunityAttachedToRemainingJoinedCommunity(
                    deltaModularityForOtherCommunity.communityId, idOfCommunityToBeRemoved, idOfCommunityToBeExpanded);
        }

        deltaModularityTrees.remove(idOfCommunityToBeRemoved);
        deltaModularityTrees.get(idOfCommunityToBeExpanded).remove(idOfCommunityToBeRemoved);
    }

    private void updateDeltaModularityTreesForCommunityAttachedToBothJoinedCommunities(long idOfAttachedCommunity,
            long idOfCommunityToBeRemoved, long idOfCommunityToBeExpanded) {
        double newDeltaModularity = deltaModularityTrees.get(idOfCommunityToBeRemoved).get(idOfAttachedCommunity).deltaModularity
                + deltaModularityTrees.get(idOfCommunityToBeExpanded).get(idOfAttachedCommunity).deltaModularity;
        deltaModularityTrees.get(idOfAttachedCommunity).get(idOfCommunityToBeExpanded).deltaModularity = newDeltaModularity;
        deltaModularityTrees.get(idOfAttachedCommunity).remove(idOfCommunityToBeRemoved);
        deltaModularityTrees.get(idOfCommunityToBeExpanded).get(idOfAttachedCommunity).deltaModularity = newDeltaModularity;
    }

    private void updateDeltaModularityTreesForCommunityAttachedToRemovedJoinedCommunity(long idOfAttachedCommunity,
            long idOfCommunityToBeRemoved, long idOfCommunityToBeExpanded) {
        double newDeltaModularity = deltaModularityTrees.get(idOfCommunityToBeRemoved).get(idOfAttachedCommunity).deltaModularity
                - 2
                * fractionOfEdgeEndsAttachedToCommunity.get(idOfAttachedCommunity)
                * fractionOfEdgeEndsAttachedToCommunity.get(idOfCommunityToBeExpanded);
        deltaModularityTrees.get(idOfAttachedCommunity).remove(idOfCommunityToBeRemoved);
        deltaModularityTrees.get(idOfAttachedCommunity).put(idOfCommunityToBeExpanded,
                new DeltaModularity(idOfCommunityToBeExpanded, newDeltaModularity));
        deltaModularityTrees.get(idOfCommunityToBeExpanded).put(idOfAttachedCommunity,
                new DeltaModularity(idOfAttachedCommunity, newDeltaModularity));
    }

    private void updateDeltaModularityTreesForCommunityAttachedToRemainingJoinedCommunity(long idOfAttachedCommunity,
            long idOfCommunityToBeRemoved, long idOfCommunityToBeExpanded) {
        double newDeltaModularity = deltaModularityTrees.get(idOfCommunityToBeExpanded).get(idOfAttachedCommunity).deltaModularity
                - 2
                * fractionOfEdgeEndsAttachedToCommunity.get(idOfAttachedCommunity)
                * fractionOfEdgeEndsAttachedToCommunity.get(idOfCommunityToBeRemoved);
        deltaModularityTrees.get(idOfAttachedCommunity).get(idOfCommunityToBeExpanded).deltaModularity = newDeltaModularity;
        deltaModularityTrees.get(idOfCommunityToBeExpanded).get(idOfAttachedCommunity).deltaModularity = newDeltaModularity;
    }

    public Map<Long, DeltaModularity> getRow(long nodeId) {
        return deltaModularityTrees.get(nodeId);
    }

    public Set<Long> getAllRows() {
        return deltaModularityTrees.keySet();
    }

    public DeltaModularity getMaxDeltaModularityForRow(long communityId) {
        Queue<DeltaModularity> orderedDeltaModularities = deltaModularityMaxHeaps.get(communityId);
        return (orderedDeltaModularities != null) ? orderedDeltaModularities.peek() : null;
    }

    public class DeltaModularity {
        private long communityId;
        private double deltaModularity;

        public DeltaModularity(long communityId, double deltaModularity) {
            this.communityId = communityId;
            this.deltaModularity = deltaModularity;
        }

        @Override
        public String toString() {
            return "dQ:" + communityId + ":" + deltaModularity;
        }

        public long getCommunityId() {
            return communityId;
        }

        public double getDeltaModularity() {
            return deltaModularity;
        }

        public void setDeltaModularity(double deltaModularity) {
            this.deltaModularity = deltaModularity;
        }
    }

    protected Map<Long, Double> getFractionOfEdgeEndsAttachedToCommunity() {
        return fractionOfEdgeEndsAttachedToCommunity;
    }

    public void initFractionOfEdgeEndsAttachedToCommunity(long totalEdgesInGraph) {
        for (long nodeId : deltaModularityTrees.keySet()) {
            fractionOfEdgeEndsAttachedToCommunity.put(nodeId,
                    Initializer.getInitialFractionOfEdges(deltaModularityTrees.get(nodeId).size(), totalEdgesInGraph));
        }
    }
}
