package com.lumberyard.graph.neo4j.plugins.communityclustering;

import static org.junit.Assert.*;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.neo4j.graphdb.GraphDatabaseService;
import org.neo4j.test.TestGraphDatabaseFactory;

public class GraphImporterTest {
    private GraphImporter importer = new GraphImporter();
    private GraphDatabaseService graphDb;

    @Before
    public void setupDb() {
        graphDb = new TestGraphDatabaseFactory().newImpermanentDatabaseBuilder().newGraphDatabase();
    }

    @After
    public void shutDownDb() {
        graphDb.shutdown();
    }

    @Test
    public void testCreateDeltaModularityMatrixFromNeo4jGraph_allNodesCreatedAsRowsInMatrix() {
        DeltaModularityMatrix q = importer.initDeltaModularityMatrixFrom(graphDb);

        assertEquals(8, q.getAllRows().size());
    }

    @Test
    public void testCreateDeltaModularityMatrixFromNeo4jGraph_n7n8ReturnedAsBestJoin() {
        DeltaModularityMatrix q = importer.initDeltaModularityMatrixFrom(graphDb);
        MaxDeltaModularityHeap maxHeap = new MaxDeltaModularityHeap(q);

        assertEquals(Neo4jTestGraph.NODE7_NAME, graphDb.getNodeById(maxHeap.getMaxDeltaModularityJoin().firstCommunity)
                .getProperty("name"));
        assertEquals(Neo4jTestGraph.NODE8_NAME, graphDb.getNodeById(maxHeap.getMaxDeltaModularityJoin().secondCommunity)
                .getProperty("name"));
    }
}
